import os
import pytest


@pytest.fixture()
def test_data_dir() -> str:
    """Test data directory that stores some files for testing purposes.

    Returns:
        str: Test data directory.
    """
    tests_dir_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(tests_dir_path, "..", "data")



import numpy as np
import musher


def test_hpcp():
    tone = 100.
    frequencies = [tone, tone * 2, tone * 3, tone * 4]
    magnitudes = [1., 1., 1., 1.]

    harmonics = 3
    band_preset = False
    min_frequency = 50.0
    max_frequency = 500.0

    actual_hpcp = musher.hpcp(frequencies,
                              magnitudes,
                              harmonics=harmonics,
                              band_preset=band_preset,
                              min_frequency=min_frequency,
                              max_frequency=max_frequency)

    expected_hpcp = [0.,
                     0.,
                     0.,
                     0.13404962,
                     0.,
                     0.24760914,
                     0.,
                     0.,
                     0.,
                     0.,
                     1.,
                     0.]

    assert np.allclose(actual_hpcp, expected_hpcp, rtol=1e-8)


def test_hpcp_from_peaks():
    buffer = [0.] * (400 + 1)
    buffer[100] = 1.
    buffer[200] = 1.
    buffer[300] = 1.
    buffer[400] = 1.

    harmonics = 3
    band_preset = False
    min_frequency = 50.0
    max_frequency = 500.0

    spectral_peaks = musher.spectral_peaks(buffer, sample_rate=0)
    actual_hpcp = musher.hpcp_from_peaks(spectral_peaks,
                                         harmonics=harmonics,
                                         band_preset=band_preset,
                                         min_frequency=min_frequency,
                                         max_frequency=max_frequency)

    expected_hpcp = [0.,
                     0.,
                     0.,
                     0.13404962,
                     0.,
                     0.24760914,
                     0.,
                     0.,
                     0.,
                     0.,
                     1.,
                     0.]

    assert np.allclose(actual_hpcp, expected_hpcp, rtol=1e-8)



import os
import math

import numpy as np
import musher


# WAV

def test_decode_wav_from_data(test_data_dir: str):
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "mozart_c_major_30sec.wav")
    file_data = musher.load_audio_file(audio_file_path)
    actual_decoded_wav = musher.decode_wav_from_data(file_data)
    actual_normalized_samples_sum = np.sum(
        actual_decoded_wav["normalized_samples"])

    expected_decoded_wav = {
        'avg_bitrate_kbps': 1411,
        'bit_depth': 16,
        'channels': 2,
        'file_type': 'wav',
        'length_in_seconds': 30.0,
        'mono': False,
        'sample_rate': 44100,
        'samples_per_channel': 1323000,
        'stereo': True
    }
    expected_normalized_samples_sum = -996.6949157714844

    # We are not checking normalized_samples.
    assert all(item in actual_decoded_wav.items()
               for item in expected_decoded_wav.items())

    assert actual_normalized_samples_sum == expected_normalized_samples_sum


def test_decode_wav_from_file(test_data_dir: str):
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "mozart_c_major_30sec.wav")
    actual_decoded_wav = musher.decode_wav_from_file(audio_file_path)
    actual_normalized_samples_sum = np.sum(
        actual_decoded_wav["normalized_samples"])

    expected_decoded_wav = {
        'avg_bitrate_kbps': 1411,
        'bit_depth': 16,
        'channels': 2,
        'file_type': 'wav',
        'length_in_seconds': 30.0,
        'mono': False,
        'sample_rate': 44100,
        'samples_per_channel': 1323000,
        'stereo': True
    }
    expected_normalized_samples_sum = -996.6949157714844

    # We are not checking normalized_samples.
    assert all(item in actual_decoded_wav.items()
               for item in expected_decoded_wav.items())

    assert actual_normalized_samples_sum == expected_normalized_samples_sum


# MP3

def test_decode_mp3_from_file(test_data_dir: str):
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "mozart_c_major_30sec.mp3")
    actual_decoded_mp3 = musher.decode_mp3_from_file(audio_file_path)
    actual_normalized_samples_sum = np.sum(
        actual_decoded_mp3["normalized_samples"])

    expected_decoded_mp3 = {
        'avg_bitrate_kbps': 320,
        'channels': 2,
        'file_type': 'mp3',
        'length_in_seconds': 30.066938775510206,
        'mono': False,
        'sample_rate': 44100,
        'samples_per_channel': 1325952,
        'stereo': True
    }

    # linux_i686 produces a slightly different result.
    expected_normalized_samples_sum = -32665783.0
    expected_normalized_samples_sum_linux_i686 = -32665751.0

    # We are not checking normalized_samples.
    assert all(item in actual_decoded_mp3.items()
               for item in expected_decoded_mp3.items())

    assert actual_normalized_samples_sum in (
        expected_normalized_samples_sum, expected_normalized_samples_sum_linux_i686)


# OTHERS

def test_load_audio_file(test_data_dir: str):
    abs_audio_file_path = os.path.join(
        test_data_dir, "audio_files", "impulses_1second_44100.wav")
    file_data = musher.load_audio_file(abs_audio_file_path)

    actual_file_data_len = len(file_data)
    actual_file_data_sum = sum(file_data)

    assert actual_file_data_len == 882044
    assert actual_file_data_sum == 5774



import os

import musher


def test_mono_mixer(test_data_dir: str):
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "impulses_1second_44100.wav")

    wav_decoded = musher.decode_wav_from_file(audio_file_path)
    mixed_audio = musher.mono_mixer(wav_decoded["normalized_samples"])

    actual_mixed_audio_len = len(mixed_audio)
    actual_mixed_audio_sum = sum(mixed_audio)

    expected_mixed_audio_len = 441000
    expected_mixed_audio_sum = 8.999725341796875

    assert actual_mixed_audio_len == expected_mixed_audio_len
    assert actual_mixed_audio_sum == expected_mixed_audio_sum



import musher


def test_peak_detect():
    r"""
    Visual of input:
                       /\
                      /  \
               /\    /    \
              /  \  /      \
          /\ /    \/        \
        _/                   \_
    """
    inp = [0., 2., 1., 4., 1., 6., 0.]
    #          ^       ^       ^
    # ^ = peak
    #
    actual_peaks = musher.peak_detect(inp, interpolate=False)

    expected_peaks = [
        # (position, height)
        (1.0, 2.0),
        (3.0, 4.0),
        (5.0, 6.0)
    ]

    assert list(actual_peaks) == expected_peaks



from typing import List

import numpy as np
import musher


def normalize(window: List[float]) -> List[float]:
    return [2.0*i/sum(window) for i in window]


def triangular(window: List[float]) -> List[float]:
    size = len(window)
    return [2.0/size*(size/2.0-abs(float(i-(size-1.0)/2.0))) for i in range(size)]


def test_windowing_square():
    inp = [1.] * 1024
    actual_window = musher.windowing(
        inp, musher.square, zero_padding_size=0, zero_phase=False)

    expected_window = normalize(inp)
    assert np.array_equal(actual_window, expected_window)


def test_windowing_custom_triangular():
    """Example of using a custom windowing function (triangular window).

    Custom functions must accept a list of floats and return a list of floats.
    """
    inp = [1.] * 1024
    actual_window = musher.windowing(
        inp, triangular, zero_padding_size=0, zero_phase=False)

    expected_window = normalize(triangular(inp))
    assert np.array_equal(actual_window, expected_window)



import musher


def test_framecutter():
    buffer = [1, 2, 3, 4, 5, 6, 7]
    frame_size = 4
    hop_size = 2
    start_from_center = False
    last_frame_to_end_of_file = True
    valid_frame_threshold_ratio = 0.

    actual_framecutter = musher.Framecutter(buffer,
                                            frame_size,
                                            hop_size,
                                            start_from_center,
                                            last_frame_to_end_of_file,
                                            valid_frame_threshold_ratio)

    expacted_framecutter = [
        [1.0, 2.0, 3.0, 4.0],
        [3.0, 4.0, 5.0, 6.0],
        [5.0, 6.0, 7.0, 0.0],
        [7.0, 0.0, 0.0, 0.0]
    ]

    assert list(actual_framecutter) == expacted_framecutter



import numpy as np
import musher


def test_spectrum():
    inp_size = 100
    inp = [1.] * inp_size
    actual_spectrum = musher.convert_to_frequency_spectrum(inp)

    expected_spectrum = [100.] + [0.] * int((inp_size / 2))

    assert np.array_equal(actual_spectrum, expected_spectrum)



import os
import math

import musher


def test_detect_key(test_data_dir: str):
    """Detect the key of a music file.
    """
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "mozart_c_major_30sec.wav")
    wav_decoded = musher.decode_wav_from_file(audio_file_path)
    normalized_samples = wav_decoded["normalized_samples"]
    sample_rate = wav_decoded["sample_rate"]

    actual_key_output = musher.detect_key(
        normalized_samples, sample_rate, "Temperley")

    expected_key_output = {
        'key': 'C',
        'scale': 'major',
        'strength': 0.760322,
        'first_to_second_relative_strength': 0.607807
    }

    assert actual_key_output['key'] == expected_key_output['key']
    assert actual_key_output['scale'] == expected_key_output['scale']
    assert math.isclose(actual_key_output['strength'],
                        expected_key_output['strength'], rel_tol=1e-6)
    assert math.isclose(actual_key_output['first_to_second_relative_strength'],
                        expected_key_output['first_to_second_relative_strength'], rel_tol=1e-6)


def test_estimate_key(test_data_dir: str):
    """Manually performing the detect_key function.

    Performance is slower than detect_key because of multiple copies of large
    arrays between C++ and Python.
    """
    audio_file_path = os.path.join(
        test_data_dir, "audio_files", "mozart_c_major_30sec.wav")
    pcp_size = 36
    sample_rate = 44100.
    num_harmonics = 4
    wav_decoded = musher.decode_wav_from_file(audio_file_path)
    mixed_audio = musher.mono_mixer(wav_decoded["normalized_samples"])
    count = 0
    sums = [0] * pcp_size
    for frame in musher.Framecutter(mixed_audio, frame_size=4096, hop_size=512):
        windowed_frame = musher.windowing(frame, musher.blackmanharris62dB)
        spectrum = musher.convert_to_frequency_spectrum(windowed_frame)
        spectral_peaks = musher.spectral_peaks(spectrum,
                                               sort_by="height",
                                               sample_rate=sample_rate,
                                               max_pos=int(sample_rate / 2))
        hpcp = musher.hpcp_from_peaks(spectral_peaks,
                                      pcp_size,
                                      harmonics=num_harmonics - 1,
                                      window_size=.5)
        for idx, _ in enumerate(hpcp):
            sums[idx] += hpcp[idx]
        count += 1

    avgs = [x/count for x in sums]

    actual_key_output = musher.estimate_key(avgs, profile_type="Temperley")

    expected_key_output = {
        'key': 'C',
        'scale': 'major',
        'strength': 0.7603224296919142,
        'first_to_second_relative_strength': 0.6078072169442225
    }

    assert actual_key_output['key'] == expected_key_output['key']
    assert actual_key_output['scale'] == expected_key_output['scale']
    assert math.isclose(actual_key_output['strength'],
                        expected_key_output['strength'], rel_tol=1e-6)
    assert math.isclose(actual_key_output['first_to_second_relative_strength'],
                        expected_key_output['first_to_second_relative_strength'], rel_tol=1e-6)



"""Setup that is responsible for building the C++ python extensions.
"""
import os
import platform
import subprocess
import shutil
import glob
import signal
from typing import Callable
from setuptools import setup, find_packages, Extension, Command
from setuptools.command.test import test

import pybind11


ROOT_DIR = os.path.dirname(os.path.realpath(__file__))


def get_build_dir() -> str:
    """Get the build directory that will store all C/C++ extension by-products.

    Returns:
        str: build directory path.
    """
    return os.path.join(ROOT_DIR, "build")


# pylint: disable=no-self-use
# pylint: disable=unnecessary-pass
# pylint: disable=attribute-defined-outside-init
class CleanProject(Command):
    """Clean the project directory of temporary files involved with
    building the C extension and python module.
    This will also uninstall the development version of musher.
    """
    user_options = []

    def initialize_options(self):
        """Initialize user options
        """
        pass

    def finalize_options(self):
        """Finalize user options
        """
        pass

    def run(self):
        """Run the clean
        """

        cleanup_list = [
            # Cmake
            os.path.join(ROOT_DIR, "build"),
            # Python
            os.path.join(ROOT_DIR, "dist"),
            os.path.join(ROOT_DIR, "musher.egg-info"),
            os.path.join(ROOT_DIR, ".eggs"),
            os.path.join(ROOT_DIR, ".pytest_cache"),
            os.path.join(ROOT_DIR, ".tox"),
            *glob.glob(os.path.join(ROOT_DIR, "musher", "*.so")),
            # Doxygen
            os.path.join(ROOT_DIR, "docs", "html"),
            os.path.join(ROOT_DIR, "docs", "latex"),
            os.path.join(ROOT_DIR, "docs", "xml"),
            # Sphinx
            os.path.join(ROOT_DIR, "docs", "build")
        ]

        for item in cleanup_list:
            try:  # If item is a dir then remove it
                shutil.rmtree(item)
                print("deleted {}".format(item))
            except OSError:
                try:  # If item is a file then remove it
                    os.remove(item)
                    print("deleted {}".format(item))
                except OSError:
                    pass

        print(u'\u2713', "cleaning done")


class PublishDocs(Command):
    """Publish the generated documentation to Github pages

        You must install the python module and generate the docs BEFORE running this command.
            1. pip install -e .
            2. python setup.py cmake --docs

        Example usage:
            python setup.py publish_docs -m "Added docs for new function"

        Helpful reads:
            https://discourse.gohugo.io/t/simple-deployment-to-gh-pages/5003
    """
    user_options = [
        ('message=', 'm', "Commit message when publishing docs."),
    ]

    def initialize_options(self):
        """Initialize user options
        """
        self.message = "Updated docs"

    def finalize_options(self):
        """Finalize user options
        """
        pass

    def copytree(self, src: str, dst: str, ignore: Callable = None):
        """copytree for older python versions that don't have the latest
        version of shutil.copytree(dirs_exist_ok=True)
        """
        if os.path.isdir(src):
            if not os.path.isdir(dst):
                os.makedirs(dst)
            files = os.listdir(src)
            if ignore is not None:
                ignored = ignore(src, files)
            else:
                ignored = set()
            for _file in files:
                if _file not in ignored:
                    self.copytree(os.path.join(src, _file),
                                  os.path.join(dst, _file),
                                  ignore)
        else:
            shutil.copyfile(src, dst)

    def run(self):
        """Publish to github pages
        """

        sphinx_build_dir = os.path.join(ROOT_DIR, 'build', 'docs', 'sphinx')
        temp_gh_pages_dir = os.path.join(ROOT_DIR, 'temp_gh_pages')
        git_worktree = os.path.join(
            ROOT_DIR, '.git', 'worktrees', "temp_gh_pages")
        nojekyll = os.path.join(temp_gh_pages_dir, ".nojekyll")

        # Cleanup to ensure we are starting fresh.
        try:
            shutil.rmtree(temp_gh_pages_dir)
        except OSError:
            pass
        subprocess.run(['git', 'worktree', 'prune'], cwd=ROOT_DIR, check=True)
        try:
            shutil.rmtree(git_worktree)
        except OSError:
            pass

        print("\nAdding gh-pages branch as a worktree...")
        subprocess.run(['git', 'worktree', 'add', '-B', 'gh-pages',
                        temp_gh_pages_dir, 'origin/gh-pages'], cwd=ROOT_DIR, check=True)
        # Pull to ensure we are up to date.
        subprocess.run(['git', 'pull'], cwd=temp_gh_pages_dir, check=True)

        # Copy newly generated docs
        print("\nCopying newly generated docs to git branch...")
        try:
            shutil.copytree(sphinx_build_dir, temp_gh_pages_dir,
                            dirs_exist_ok=True)
        except TypeError:
            # Not using python3.8+
            self.copytree(sphinx_build_dir, temp_gh_pages_dir)

        # Add a .nojekyll files to ensure gh-pages uses index.html as the root of the site.
        open(nojekyll, 'w').close()

        subprocess.run(['git', 'add', '--all'],
                       cwd=temp_gh_pages_dir, check=True)
        try:
            print("\nCommitting changes...")
            subprocess.run(['git', 'commit', '-m', self.message],
                           cwd=temp_gh_pages_dir, check=True)
            print("\nPushing to github...")
            subprocess.run(['git', 'push', 'origin', 'gh-pages'],
                           cwd=temp_gh_pages_dir, check=True)
        except subprocess.CalledProcessError as err:
            if err.returncode != 1:
                # returncode == 1 means the branch is up to date.
                raise err

        shutil.rmtree(temp_gh_pages_dir)

        subprocess.run(['git', 'worktree', 'prune'], cwd=ROOT_DIR, check=True)


class CMakeBuild(Command):
    """
    Debug (with tests): python setup.py cmake --debug
    Release (no tests): python setup.py cmake

    Generate docs: python setup.py cmake --docs
    """
    description = 'Build the C++ code with various options.'
    user_options = [
        ('debug', 'd', "Compile in debug mode."),
        ('docs', 'r', "Generate documenation for C++ and Python."),
    ]

    def initialize_options(self):
        """Initialize user options
        """
        self.debug = False
        self.docs = False

    def finalize_options(self):
        """Finalize user options
        """
        pass

    def run(self):
        """Build the C++ code.

        Raises:
            RuntimeError: Error if Cmake is not installed.
        """
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError as err:
            raise RuntimeError(
                "CMake must be installed to build musher") from err

        build_dir = get_build_dir()
        cmake_args = []

        if self.debug:
            cmake_args += ["-DCMAKE_BUILD_TYPE=Debug", "-DENABLE_TESTS=On"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=Release"]

        if self.docs:
            cmake_args += ["-DGENERATE_DOCS=On"]
        else:
            cmake_args += ["-DGENERATE_DOCS=Off"]

        if not os.path.exists(build_dir):
            os.makedirs(build_dir)

        subprocess.run(['cmake', ROOT_DIR] + cmake_args,
                       cwd=build_dir,
                       check=True)
        subprocess.run(['cmake', '--build', '.'],
                       cwd=build_dir,
                       check=True)


class CTest(test):
    """Run tests if compiled (only through DEBUG)

    python setup.py cmake --debug
    python setup.py ctest
    """

    def run(self):
        build_dir = get_build_dir()

        if platform.system().lower() == "windows":
            build_dir_win_debug = os.path.join(build_dir, "Debug")
            if os.path.isdir(build_dir_win_debug):
                build_dir = build_dir_win_debug

        try:
            result = subprocess.run(['ctest', "--output-on-failure"],
                                    cwd=build_dir,
                                    check=True,
                                    stderr=subprocess.PIPE)
        except FileNotFoundError:
            # cwd was not found.
            print("HINT: Did you compile the code first? "
                  "(python setup.py cmake --debug)")
            raise

        if result.stderr:
            print(result.stderr.decode("utf-8"))
            print("HINT: Did you compile the code in debug mode first? "
                  "(python setup.py cmake --debug)")

        if result.returncode == -signal.SIGSEGV:
            print("C++ Seg fault.")


class GTest(test):
    r"""Run tests if compiled (only through DEBUG)

    python setup.py cmake --debug
    python setup.py gtest

    example using filters:
        python setup.py gtest --g=--gtest_filter=HPCP.\*

    Attributes:
        gtest_options (str): Any gtest options that should be used when running the tests.

    """
    description = 'Run google tests for c++ library'
    user_options = [
        ('gtest-options=', 'g',
         'Any google test args that need to be passed through when building tests.'),
    ]

    def initialize_options(self):
        self.gtest_options = ""

    def finalize_options(self):
        pass

    def run(self):
        bin_dir = os.path.join(get_build_dir(), "bin")

        if platform.system().lower() == "windows":
            bin_dir_win_debug = os.path.join(bin_dir, "Debug")
            if os.path.isdir(bin_dir_win_debug):
                bin_dir = bin_dir_win_debug

        try:
            result = subprocess.run(['./musher-core-test', self.gtest_options],
                                    cwd=bin_dir,
                                    check=True)
        except FileNotFoundError:
            # Test executable not found / cwd not found.
            print("HINT: Did you compile the code in debug mode first? "
                  "(python setup.py cmake --debug)")
            raise

        if result.returncode == -11:
            print("C++ Seg fault.")


def extra_compile_args() -> list:
    """Platform dependent extras

    Returns:
        list: Extra compile arguments
    """
    args = []
    if platform.system() == 'Darwin':
        # Something with OS X Mojave causes libstd not to be found
        args += ['-mmacosx-version-min=10.12']

    if os.name != 'nt':
        if platform.machine() == 'i686':
            # This makes GCC generate modern SSE2 instructions
            # that give the exact IEEE 754 floating-point semantics.
            # Basically ensures that calculations using produce the
            # same results on all 32bit linux architecture.
            args += ['-msse2']
        args += ['-std=c++14']

    return args


def extra_link_args() -> list:
    """Platform dependent extras

    Returns:
        list: Extra link arguments
    """
    args = []
    if platform.system() == 'Darwin':
        # Something with OS X Mojave causes libstd not to be found
        args += ['-stdlib=libc++', '-mmacosx-version-min=10.12']

    return args


setup(
    ext_modules=[
         Extension(
             'musher.musher_python',  # Destination of .so
             include_dirs=[
                # https://caligari.dartmouth.edu/doc/ibmcxx/en_US/doc/complink/tasks/tuinclud.htm
                # Allows for root level imports within C++
                ROOT_DIR,
                # 3rd party libraries
                "src/third-party",
                pybind11.get_include()
             ],
             sources=[
                 'src/python/module.cpp',
                 'src/python/wrapper.cpp',
                 'src/python/utils.cpp',
                 'src/core/audio_decoders.cpp',
                 'src/core/utils.cpp',
                 'src/core/key.cpp',
                 'src/core/hpcp.cpp',
                 'src/core/framecutter.cpp',
                 'src/core/windowing.cpp',
                 'src/core/peak_detect.cpp',
                 'src/core/spectral_peaks.cpp',
                 'src/core/spectrum.cpp',
                 'src/core/mono_mixer.cpp'
             ],
             depends=[
                 'src/python/module.h',
                 'src/python/wrapper.h',
                 'src/python/utils.h',
                 'src/core/audio_decoders.h',
                 'src/core/utils.h',
                 'src/core/key.h'
                 'src/core/hpcp.h',
                 'src/core/framecutter.h',
                 'src/core/windowing.h',
                 'src/core/peak_detect.h',
                 'src/core/spectral_peaks.h',
                 'src/core/spectrum.h',
                 'src/core/mono_mixer.h'
             ],
             extra_compile_args=extra_compile_args(),
             extra_link_args=extra_link_args(),
         )
    ],
    cmdclass={
        "cmake": CMakeBuild,
        "ctest": CTest,
        "gtest": GTest,
        "clean": CleanProject,
        "publish_docs": PublishDocs
    },
    setup_requires=['wheel', 'cython', 'setuptools_scm>=4.1.2', 'pybind11>=2.6.0', 'numpy>=1.18.5'],
    use_scm_version=True
)



from conans import ConanFile, CMake

class MusherConan(ConanFile):
    name = "musher"
    settings = "os", "compiler", "build_type", "arch"

    generators = "cmake"

    _cmake = None

    def requirements(self):
        self.requires("gtest/[>=1.10.0]")
        # self.requires("functionalplus/v0.2.10-p0@dobiasd/stable")

    def set_version(self):
        self.version = "0.0.1"

    def _configure_cmake(self):
        if self._cmake:
            return self._cmake

        cmake = CMake(self)
        cmake.definitions["CMAKE_BUILD_TYPE"] = self.settings.build_type
        cmake.definitions["ENABLE_CONAN_INSTALL"] = False
        cmake.definitions["ENABLE_PACKAGE_BUILD"] = True
        cmake.configure()
        self._cmake = cmake
        return cmake

    def build(self):
        cmake = self._configure_cmake()
        cmake.build()

    def package(self):
        cmake = self._configure_cmake()
        cmake.install()

    def package_info(self):
        self.cpp_info.libs = []

    def imports(self):
        self.copy("*.dll", dst="bin", src="bin")
